# pyright: reportUnannotatedClassAttribute=false

# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Defines the core VariableDescriptor class and predefined quantities.

This module provides a structured, extensible way to represent variables based on
physical quantities (e.g., pressure, velocity) independently of any product-specific
naming conventions.
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Generic, Literal

from typing_extensions import TypeVar

from ansys.units.base_dimensions import BaseDimensions
from ansys.units.dimensions import Dimensions
from ansys.units.quantity_dimensions import QuantityDimensions


class QuantityKind(Enum):
    """Enumeration for :class:`VariableDescriptor`."""

    SCALAR = "scalar"
    VECTOR = "vector"


QuantityKindT = TypeVar(
    "QuantityKindT", bound=QuantityKind, default=QuantityKind, covariant=True
)


@dataclass(frozen=True, unsafe_hash=True)
class VariableDescriptor(Generic[QuantityKindT]):
    """Defines a physical quantity variable descriptor."""

    dimension: Dimensions
    name: str = field(init=False)

    def __set_name__(self, _, name: str) -> None:
        object.__setattr__(self, "name", _validate_and_transform_variable(name))


ScalarVariableDescriptor = VariableDescriptor[
    Literal[QuantityKind.SCALAR]
]  #: Type alias for scalar variables
VectorVariableDescriptor = VariableDescriptor[
    Literal[QuantityKind.VECTOR]
]  #: Type alias for vector variables


def _validate_and_transform_variable(variable: str) -> str:
    """
    Validate and transform a variable name.

    Ensures the variable name is uppercase and returns its lowercase equivalent.

    Parameters
    ----------
    variable : str
        The name of the variable.

    Returns
    -------
    str
        The lowercase equivalent of the variable name.

    Raises
    ------
    ValueError
        If the variable name is not uppercase.
    """
    if not variable.isupper():
        raise ValueError(
            f"Variable names in VariableCatalog must be uppercase. "
            f"Invalid name: '{variable}'."
        )
    return variable.lower()


_D = QuantityDimensions
_B = BaseDimensions
_Q = QuantityKind


class VariableCatalogBase:
    """A catalog of variable descriptors."""

    @classmethod
    def all(cls) -> dict[str, list[VariableDescriptor]]:
        """
        Return all defined
        :class:`~ansys.units.variable_descriptor.variable_descriptor.VariableDescriptor`
        objects, organized by subcategory.

        Returns
        -------
        dict[str, list[VariableDescriptor]]
            A dictionary where keys are subcategory names (or "main" for the top-level catalog)
            and values are lists of :class:`~ansys.units.variable_descriptor.variable_descriptor.VariableDescriptor` instances.
        """  # noqa: E501
        result = {"main": []}

        for key, value in cls.__dict__.items():
            if isinstance(value, VariableDescriptor):
                result["main"].append(value)
            elif (
                isinstance(value, type)
                and issubclass(value, object)
                and value is not object
            ):
                # Collect descriptors from subcategories
                subcategory_name = key
                result[subcategory_name] = [
                    v
                    for k, v in value.__dict__.items()
                    if isinstance(v, VariableDescriptor)
                ]

        return result

    @classmethod
    def add(
        cls, variable: str, dimension: Dimensions, subcategory: str | None = None
    ) -> None:
        """
        Add a variable to the catalog.

        Parameters
        ----------
        variable : str
            The name of the variable (must be uppercase).
        dimension : Dimensions
            The dimension of the variable.
        subcategory: str|None
            The optional subcategory for the variable.

        Raises
        ------
        ValueError
            The variable name is not uppercase or already exists.
        """
        _validate_and_transform_variable(
            variable
        )  # Validate variable name before adding to prevent invalid state

        # Determine the target category (main catalog or subcategory)
        target = cls
        if subcategory:
            # Ensure the subcategory exists as an attribute
            if not hasattr(cls, subcategory):
                # Create the subcategory and assign a docstring
                subcategory_class = type(
                    subcategory,
                    (object,),
                    {
                        "__doc__": (
                            f"Dictionary of variable descriptors for {subcategory}-related "
                            f"quantities."
                        )
                    },
                )
                setattr(cls, subcategory, subcategory_class)
            target = getattr(cls, subcategory)

        # Check if the variable already exists in the target category
        if hasattr(target, variable):
            raise ValueError(
                f"Variable name '{variable}' already exists in the '{subcategory or 'main'}' "
                "catalog. Please choose a unique name."
            )

        # Add the variable to the target category
        descriptor = VariableDescriptor(dimension)
        descriptor.__set_name__(target, variable)
        setattr(target, variable, descriptor)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ``UnitRegistry`` class and instance-scoped unit registration."""

from collections.abc import Generator, Mapping
import math
import os
from typing import TYPE_CHECKING, Any

import yaml

from ansys.units._constants import (
    _BaseUnitInfo,
)
from ansys.units._constants import (
    _DerivedUnitInfo,
)
from ansys.units._constants import _base_units as _CONST_BASE_UNITS
from ansys.units._constants import _derived_units as _CONST_DERIVED_UNITS
from ansys.units.unit import Unit


class UnitRegistry:
    """
    A representation of valid ``Unit`` instances.

    All base and derived units loaded from the configuration file, `cfg.yaml`,
    on package initialization are provided by default.

    Parameters
    ----------
    config: str, optional
        Path of a ``YAML`` configuration file, which can be a custom file, and
        defaults to the provided file, ``cfg.yaml``. Custom configuration files
        must match the format of the default configuration file.
    other: dict, optional
        Dictionary for additional units.

    Examples
    --------
    >>> from ansys.units import UnitRegistry, Unit
    >>> ureg = UnitRegistry()
    >>> assert ureg.kg == Unit(units="kg")
    >>> fps = Unit("ft s^-1")
    >>> ureg.foot_per_sec = fps
    """

    def __init__(
        self,
        config: str = "cfg.yaml",
        other: Mapping[
            str, Mapping[str, Any]
        ] = {},  # pyright: ignore[reportCallInDefaultInitializer]
    ):
        unitdict = dict(other)

        if config:
            file_dir = os.path.dirname(__file__)
            qc_path = os.path.join(file_dir, config)

            with open(qc_path, "r") as qc_yaml:
                qc_data = yaml.safe_load(qc_yaml)
                _base_units: dict[str, _BaseUnitInfo] = qc_data["base_units"]
                _derived_units: dict[str, _DerivedUnitInfo] = qc_data["derived_units"]

            unitdict |= _base_units | _derived_units

        for unit_name in unitdict:
            # Prevent overriding attributes already present on this instance
            if hasattr(self, unit_name):
                raise UnitAlreadyRegistered(unit_name)

            cfg = unitdict[unit_name]
            if unit_name in _CONST_BASE_UNITS or unit_name in _CONST_DERIVED_UNITS:
                object.__setattr__(self, unit_name, Unit(unit_name, cfg))
            else:
                # For dynamically registered units not present in constants, build
                # from their composition so dimensions/si data are correct, then
                # override the name to the desired symbol and attach config.
                if "composition" in cfg:
                    composed = Unit(units=str(cfg["composition"]))
                    obj = Unit(copy_from=composed)
                    obj._name = unit_name
                    object.__setattr__(self, unit_name, obj)
                else:
                    object.__setattr__(self, unit_name, Unit(unit_name, cfg))

    def __str__(self):
        returned_string = ""
        attrs = self.__dict__
        for key in attrs:
            returned_string += f"{key}, "
        return returned_string

    if TYPE_CHECKING:

        def __getattr__(self, name: str) -> Unit: ...

    def __setattr__(self, name: str, unit: Any) -> None:
        if hasattr(self, name):
            raise UnitAlreadyRegistered(name)
        self.__dict__[name] = unit

    def __iter__(self) -> Generator[str]:
        yield from self.__dict__

    def register_unit(
        self,
        *,
        unit: str,
        composition: str,
        factor: float,
    ) -> Unit:
        """
        Register a new derived unit on this ``UnitRegistry`` instance.

        This is instance-scoped: it affects only this registry and does not
        mutate global state or other registries.

        Parameters
        ----------
        unit: str
            The symbol/name of the new unit (e.g., "Q"). Preferred keyword.
        composition: str
            A valid unit composition using existing configured units (e.g., "N m").
        factor: float
            Scale factor that relates the composition to this unit.

        Returns
        -------
        Unit
            The registered unit attached on this instance.

        Raises
        ------
        UnitAlreadyRegistered
            If a unit with the same name already exists on this instance or is built-in.
        ValueError
            If ``unit`` is empty or ``factor`` is not finite.
        """
        unit = unit.strip()
        if not unit:
            raise ValueError("`unit` must be a non-empty string.")
        f = float(factor)
        if not math.isfinite(f):
            raise ValueError("`factor` must be a finite number.")

        # Prevent overriding built-ins or existing attributes on this instance
        if (
            unit in _CONST_BASE_UNITS
            or unit in _CONST_DERIVED_UNITS
            or hasattr(self, unit)
        ):
            raise UnitAlreadyRegistered(unit)

        composed = Unit(units=str(composition))
        obj = Unit(copy_from=composed)

        obj._si_scaling_factor *= f
        obj._name = unit
        object.__setattr__(self, unit, obj)
        return obj


class UnitAlreadyRegistered(ValueError):
    """Raised when a unit has previously been registered."""

    def __init__(self, name: str):
        super().__init__(f"Unable to override `{name}` it has already been registered.")



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.

# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# This file is @generated

from typing import Literal

UnitKey = Literal[
    "kg",
    "g",
    "lb",
    "lbm",
    "slug",
    "m",
    "cm",
    "ft",
    "inch",
    "in",
    "s",
    "A",
    "mol",
    "slugmol",
    "cd",
    "sr",
    "radian",
    "degree",
    "K",
    "C",
    "F",
    "R",
    "delta_K",
    "delta_C",
    "delta_F",
    "delta_R",
    "N",
    "Pa",
    "W",
    "J",
    "V",
    "farad",
    "H",
    "S",
    "Wb",
    "T",
    "dyne",
    "erg",
    "h",
    "pdl",
    "psi",
    "lbf",
    "psf",
    "ohm",
    "Hz",
    "l",
    "gal",
    "BTU",
    "cal",
    "coulomb",
    "tonne",
]

MassKey = Literal[
    "kg",
    "g",
    "lb",
    "lbm",
    "slug",
]

LengthKey = Literal[
    "m",
    "cm",
    "ft",
    "inch",
    "in",
]

TimeKey = Literal["s",]

CurrentKey = Literal["A",]

ChemicalAmountKey = Literal[
    "mol",
    "slugmol",
]

LightKey = Literal["cd",]

SolidAngleKey = Literal["sr",]

AngleKey = Literal[
    "radian",
    "degree",
]

TemperatureKey = Literal[
    "K",
    "C",
    "F",
    "R",
]

TemperatureDifferenceKey = Literal[
    "delta_K",
    "delta_C",
    "delta_F",
    "delta_R",
]

QuantityKey = Literal[
    "Mass",
    "Length",
    "Time",
    "Temperature",
    "Current",
    "SubstanceAmount",
    "Light",
    "Angle",
    "SolidAngle",
    "Acceleration",
    "AngularAcceleration",
    "AngularVelocity",
    "Area",
    "Compressibility",
    "Concentration",
    "DecayConstant",
    "Density",
    "DynamicViscosity",
    "ElectricCharge",
    "ElectricChargeDensity",
    "ElectricCurrentDensity",
    "ElectricCurrentSource",
    "ElectricFluxDensity",
    "Enthalpy",
    "Force",
    "ForceDensity",
    "Frequency",
    "HeatFlux",
    "HeatGeneration",
    "HeatTransferCoefficient",
    "Impulse",
    "MagneticFieldIntensity",
    "MassFlowRate",
    "MassFlux",
    "MolarConcentration",
    "MolarEnergy",
    "MolarEnthalpy",
    "MolarEntropy",
    "MolarMass",
    "MolarVolume",
    "Moment",
    "MomentOfInertiaOfArea",
    "MomentOfInertiaOfMass",
    "Momentum",
    "Pressure",
    "ShearStrain",
    "ShearStrainRate",
    "SpecificConcentration",
    "SpecificEnergy",
    "SpecificEnthalpy",
    "SpecificEntropy",
    "SpecificFlameSurfaceDensity",
    "SpecificHeatCapacity",
    "SpecificVolume",
    "Stiffness",
    "Strain",
    "SurfaceChargeDensity",
    "SurfaceForceDensity",
    "SurfacePowerDensity",
    "SurfaceTension",
    "TemperatureGradient",
    "ThermalCapacitance",
    "ThermalConductance",
    "ThermalConductivity",
    "ThermalContactResistance",
    "ThermalExpansionCoefficient",
    "Torque",
    "Velocity",
    "Volume",
    "VolumetricFlowRate",
]

BaseUnit = Literal[
    "MASS",
    "LENGTH",
    "TIME",
    "TEMPERATURE",
    "TEMPERATURE_DIFFERENCE",
    "ANGLE",
    "CHEMICAL_AMOUNT",
    "LIGHT",
    "CURRENT",
    "SOLID_ANGLE",
]

Systems = Literal[
    "SI",
    "CGS",
    "BT",
]



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.

# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# This file is @generated

from typing_extensions import final

from ansys.units.base_dimensions import BaseDimensions as _B
from ansys.units.dimensions import Dimensions
from ansys.units.quantity_dimensions import QuantityDimensions as _D
from ansys.units.variable_descriptor.variable_descriptor import (
    ScalarVariableDescriptor,
    VariableCatalogBase,
    VectorVariableDescriptor,
)


@final
class VariableCatalog(VariableCatalogBase):
    """A catalog of variable descriptors."""

    ABSOLUTE_PRESSURE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    ABSORBANCE = ScalarVariableDescriptor(Dimensions({}))
    ABSORBED_DOSE_RATE = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -3.0})
    )
    ACCELERATION = VectorVariableDescriptor(Dimensions({_B.LENGTH: 1.0, _B.TIME: -2.0}))
    ACCELERATION_MAGNITUDE = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    ACCELERATION_X = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    ACCELERATION_Y = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    ACCELERATION_Z = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    ACOUSTICAL_ABSORPTION = ScalarVariableDescriptor(Dimensions({}))
    ACTION = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -1.0})
    )
    ANGLE = ScalarVariableDescriptor(Dimensions({_B.ANGLE: 1.0}))
    ANGULAR_ACCELERATION = ScalarVariableDescriptor(Dimensions({_B.TIME: -2.0}))
    ANGULAR_MOMENTUM = VectorVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -1.0})
    )
    ANGULAR_MOMENTUM_MAGNITUDE = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -1.0})
    )
    ANGULAR_MOMENTUM_X = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -1.0})
    )
    ANGULAR_MOMENTUM_Y = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -1.0})
    )
    ANGULAR_MOMENTUM_Z = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -1.0})
    )
    ANGULAR_VELOCITY = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    AREA = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 2.0}))
    AREA_DENSITY = ScalarVariableDescriptor(Dimensions({_B.MASS: 1.0, _B.LENGTH: -2.0}))
    ATOMIC_MASS = ScalarVariableDescriptor(Dimensions({_B.MASS: 1.0}))
    ATOMIC_NUMBER = ScalarVariableDescriptor(Dimensions({}))
    ATTENUATION_COEFFICIENT = ScalarVariableDescriptor(Dimensions({_B.LENGTH: -1.0}))
    AXIAL_VELOCITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    BIREFRINGENCE = ScalarVariableDescriptor(Dimensions({}))
    BULK_MODULUS = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    CAPACITANCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: -1.0, _B.LENGTH: -2.0, _B.TIME: 4.0, _B.CURRENT: 2.0})
    )
    CATALYTIC_EFFICIENCY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 3.0, _B.CHEMICAL_AMOUNT: -1.0, _B.TIME: -1.0})
    )
    CELL_REYNOLDS_NUMBER = ScalarVariableDescriptor(Dimensions({}))
    CHEMICAL_AMOUNT = ScalarVariableDescriptor(Dimensions({_B.CHEMICAL_AMOUNT: 1.0}))
    COEFFICIENT_OF_RESTITUTION = ScalarVariableDescriptor(Dimensions({}))
    COLOR = ScalarVariableDescriptor(Dimensions({}))
    COMPRESSIBILITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: -1.0, _B.LENGTH: 1.0, _B.TIME: 2.0})
    )
    COMPRESSIVE_STRENGTH = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    CONVECTIVE_COURANT_NUMBER = ScalarVariableDescriptor(Dimensions({}))
    CORROSION_RESISTANCE = ScalarVariableDescriptor(Dimensions({}))
    CREEP = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    CURIE_TEMPERATURE = ScalarVariableDescriptor(Dimensions({_B.TEMPERATURE: 1.0}))
    CURRENT = ScalarVariableDescriptor(Dimensions({_B.CURRENT: 1.0}))
    DENSITY = ScalarVariableDescriptor(Dimensions({_B.MASS: 1.0, _B.LENGTH: -3.0}))
    DIAMAGNETISM = ScalarVariableDescriptor(Dimensions({}))
    DIELECTRIC_CONSTANT = ScalarVariableDescriptor(Dimensions({}))
    DIELECTRIC_STRENGTH = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -3.0, _B.CURRENT: -1.0})
    )
    DUCTILITY = ScalarVariableDescriptor(Dimensions({}))
    DURABILITY = ScalarVariableDescriptor(Dimensions({}))
    DYNAMIC_PRESSURE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    DYNAMIC_VISCOSITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -1.0})
    )
    EFFECTIVE_PRANDTL_NUMBER = ScalarVariableDescriptor(Dimensions({}))
    EFFECTIVE_THERMAL_CONDUCTIVITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.MASS: 1.0, _B.TIME: -3.0, _B.TEMPERATURE: -1.0})
    )
    EFFECTIVE_VISCOSITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -1.0})
    )
    ELASTICITY = ScalarVariableDescriptor(Dimensions({}))
    ELECTRICAL_CAPACITANCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: -1.0, _B.LENGTH: -2.0, _B.TIME: 4.0, _B.CURRENT: 2.0})
    )
    ELECTRICAL_CONDUCTANCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: -1.0, _B.LENGTH: -2.0, _B.TIME: 3.0, _B.CURRENT: 2.0})
    )
    ELECTRICAL_CONDUCTIVITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: -1.0, _B.LENGTH: -3.0, _B.TIME: 3.0, _B.CURRENT: 2.0})
    )
    ELECTRICAL_IMPEDANCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 2.0, _B.TIME: -3.0, _B.CURRENT: -2.0})
    )
    ELECTRICAL_INDUCTANCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 2.0, _B.TIME: -2.0, _B.CURRENT: -2.0})
    )
    ELECTRICAL_POTENTIAL = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 2.0, _B.TIME: -3.0, _B.CURRENT: -1.0})
    )
    ELECTRICAL_REACTANCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 2.0, _B.TIME: -3.0, _B.CURRENT: -2.0})
    )
    ELECTRICAL_RESISTANCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 2.0, _B.TIME: -3.0, _B.CURRENT: -2.0})
    )
    ELECTRICAL_RESISTIVITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 3.0, _B.TIME: -3.0, _B.CURRENT: -2.0})
    )
    ELECTRIC_CHARGE = ScalarVariableDescriptor(
        Dimensions({_B.TIME: 1.0, _B.CURRENT: 1.0})
    )
    ELECTRIC_SUSCEPTIBILITY = ScalarVariableDescriptor(Dimensions({}))
    ELECTROCALORIC_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions(
            {
                _B.TEMPERATURE: 1.0,
                _B.MASS: -1.0,
                _B.LENGTH: -2.0,
                _B.TIME: 3.0,
                _B.CURRENT: 1.0,
            }
        )
    )
    ELECTROSTRICTION = ScalarVariableDescriptor(Dimensions({}))
    ELECTRO_OPTIC_EFFECT = ScalarVariableDescriptor(Dimensions({}))
    EMISSIVITY = ScalarVariableDescriptor(Dimensions({}))
    ENERGY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    ENERGY_DENSITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: -1.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    ENERGY_FLUX_DENSITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -3.0})
    )
    ENTHALPY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    ENTROPY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0, _B.TEMPERATURE: -1.0})
    )
    FATIGUE_LIMIT = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    FLEXURAL_MODULUS = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    FLEXURAL_STRENGTH = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    FORCE = VectorVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    FORCE_MAGNITUDE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    FORCE_X = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    FORCE_Y = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    FORCE_Z = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -2.0})
    )
    FRACTURE_TOUGHNESS = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -2.0})
    )
    FREQUENCY = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    FREQUENCY_DRIFT = ScalarVariableDescriptor(Dimensions({_B.TIME: -2.0}))
    FRICTION_COEFFICIENT = ScalarVariableDescriptor(Dimensions({}))
    FUEL_EFFICIENCY = ScalarVariableDescriptor(Dimensions({_B.LENGTH: -2.0}))
    HALF_LIFE = ScalarVariableDescriptor(Dimensions({_B.TIME: 1.0}))
    HALL_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 3.0, _B.CURRENT: -1.0, _B.TIME: -1.0})
    )
    HARDNESS = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    HEAT_CAPACITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0, _B.TEMPERATURE: -1.0})
    )
    HEAT_FLUX_DENSITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -3.0})
    )
    HEAT_OF_VAPORIZATION = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    HYGROSCOPY = ScalarVariableDescriptor(Dimensions({}))
    HYSTERESIS = ScalarVariableDescriptor(Dimensions({}))
    ILLUMINANCE = ScalarVariableDescriptor(
        Dimensions({_B.LIGHT: 1.0, _B.SOLID_ANGLE: 1.0, _B.LENGTH: -2.0})
    )
    INTERNAL_ENERGY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    JERK = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0, _B.TIME: -3.0}))
    KINEMATIC_VISCOSITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -1.0})
    )
    LENGTH = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0}))
    LIGHT = ScalarVariableDescriptor(Dimensions({_B.LIGHT: 1.0}))
    LINEAR_MASS_DENSITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0})
    )
    LUMINANCE = ScalarVariableDescriptor(Dimensions({_B.LIGHT: 1.0, _B.LENGTH: -2.0}))
    LUMINOSITY = ScalarVariableDescriptor(
        Dimensions({_B.LIGHT: 1.0, _B.SOLID_ANGLE: 1.0})
    )
    LUMINOUS_EXPOSURE = ScalarVariableDescriptor(
        Dimensions({_B.LIGHT: 1.0, _B.SOLID_ANGLE: 1.0, _B.LENGTH: -2.0, _B.TIME: 1.0})
    )
    LUMINOUS_FLUX = ScalarVariableDescriptor(
        Dimensions({_B.LIGHT: 1.0, _B.SOLID_ANGLE: 1.0})
    )
    MAGNETIC_FLUX = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 2.0, _B.TIME: -2.0, _B.CURRENT: -1.0})
    )
    MAGNETIC_INDUCTION = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -2.0, _B.CURRENT: -1.0})
    )
    MAGNETIC_PERMEABILITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.MASS: 1.0, _B.TIME: -2.0, _B.CURRENT: -2.0})
    )
    MAGNETIC_VECTOR_POTENTIAL = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.MASS: 1.0, _B.TIME: -2.0, _B.CURRENT: -1.0})
    )
    MAGNETOCALORIC_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions({_B.TEMPERATURE: 1.0, _B.MASS: -1.0, _B.TIME: 2.0, _B.CURRENT: 1.0})
    )
    MAGNETOELECTRIC_POLARIZABILITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: -1.0, _B.TIME: -1.0})
    )
    MAGNETORESISTANCE = ScalarVariableDescriptor(Dimensions({}))
    MAGNETOSTRICTION = ScalarVariableDescriptor(Dimensions({}))
    MAGNETOTHERMOELECTRIC_POWER = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -1.0, _B.TEMPERATURE: -1.0})
    )
    MALLEABILITY = ScalarVariableDescriptor(Dimensions({}))
    MASS = ScalarVariableDescriptor(Dimensions({_B.MASS: 1.0}))
    MASS_CONTROL = ScalarVariableDescriptor(Dimensions({_B.MASS: 1.0, _B.TIME: -3.0}))
    MASS_DIFFUSIVITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -1.0})
    )
    MASS_FLOW_RATE = ScalarVariableDescriptor(Dimensions({_B.MASS: 1.0, _B.TIME: -1.0}))
    MAXIMUM_ENERGY_PRODUCT = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: -1.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    MELTING_POINT = ScalarVariableDescriptor(Dimensions({_B.TEMPERATURE: 1.0}))
    MESH_VELOCITY = VectorVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MESH_VELOCITY_MAGNITUDE = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MESH_VELOCITY_X = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MESH_VELOCITY_Y = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MESH_VELOCITY_Z = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MOLALITY = ScalarVariableDescriptor(
        Dimensions({_B.CHEMICAL_AMOUNT: 1.0, _B.MASS: -1.0})
    )
    MOLARITY = ScalarVariableDescriptor(
        Dimensions({_B.CHEMICAL_AMOUNT: 1.0, _B.LENGTH: -3.0})
    )
    MOLAR_ENERGY = ScalarVariableDescriptor(
        Dimensions(
            {_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0, _B.CHEMICAL_AMOUNT: -1.0}
        )
    )
    MOLAR_ENTROPY = ScalarVariableDescriptor(
        Dimensions(
            {
                _B.LENGTH: 2.0,
                _B.MASS: 1.0,
                _B.TIME: -2.0,
                _B.TEMPERATURE: -1.0,
                _B.CHEMICAL_AMOUNT: -1.0,
            }
        )
    )
    MOLAR_HEAT_CAPACITY = ScalarVariableDescriptor(
        Dimensions(
            {
                _B.LENGTH: 2.0,
                _B.MASS: 1.0,
                _B.TIME: -2.0,
                _B.TEMPERATURE: -1.0,
                _B.CHEMICAL_AMOUNT: -1.0,
            }
        )
    )
    MOLAR_MASS = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.CHEMICAL_AMOUNT: -1.0})
    )
    MOLAR_VOLUME = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 3.0, _B.CHEMICAL_AMOUNT: -1.0})
    )
    MOMENTUM = VectorVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MOMENTUM_MAGNITUDE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MOMENTUM_X = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MOMENTUM_Y = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MOMENTUM_Z = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    MOMENT_OF_INERTIA = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 2.0})
    )
    NERNST_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -1.0, _B.TEMPERATURE: -1.0})
    )
    NEUTRON_CROSS_SECTION = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 2.0}))
    NORMALIZED_Q_CRITERION = ScalarVariableDescriptor(Dimensions({}))
    OPTICAL_ACTIVITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.MASS: -1.0})
    )
    PERMEABILITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.MASS: 1.0, _B.TIME: -2.0, _B.CURRENT: -2.0})
    )
    PERMITTIVITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: -1.0, _B.LENGTH: -3.0, _B.TIME: 4.0, _B.CURRENT: 2.0})
    )
    PH = ScalarVariableDescriptor(Dimensions({}))
    PHOTOELASTICITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: -1.0, _B.LENGTH: 1.0, _B.TIME: 2.0})
    )
    PHOTOSENSITIVITY = ScalarVariableDescriptor(Dimensions({}))
    PIEZOELECTRIC_CONSTANTS = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -1.0, _B.CURRENT: -1.0})
    )
    PIEZOMAGNETISM = ScalarVariableDescriptor(
        Dimensions({_B.CURRENT: -1.0, _B.LENGTH: 1.0})
    )
    PLASTICITY = ScalarVariableDescriptor(Dimensions({}))
    POISSON_RATIO = ScalarVariableDescriptor(Dimensions({}))
    POSITION = VectorVariableDescriptor(Dimensions({_B.LENGTH: 1.0}))
    POSITION_MAGNITUDE = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0}))
    POSITION_X = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0}))
    POSITION_Y = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0}))
    POSITION_Z = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0}))
    POWER = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -3.0})
    )
    PRANDTL_NUMBER = ScalarVariableDescriptor(Dimensions({}))
    PRESSURE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    PRESSURE_COEFFICIENT = ScalarVariableDescriptor(Dimensions({}))
    PRODUCTION_OF_TURBULENT_KINETIC_ENERGY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -3.0})
    )
    PYROELECTRICITY = ScalarVariableDescriptor(
        Dimensions({_B.CURRENT: 1.0, _B.LENGTH: -2.0, _B.TEMPERATURE: -1.0})
    )
    PYROMAGNETIC_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions(
            {_B.MASS: 1.0, _B.TIME: -2.0, _B.CURRENT: -1.0, _B.TEMPERATURE: -1.0}
        )
    )
    Q_CRITERION = ScalarVariableDescriptor(Dimensions({_B.TIME: -2.0}))
    RADIAL_VELOCITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    RADIANCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -3.0, _B.SOLID_ANGLE: -1.0})
    )
    RADIANT_EXPOSURE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -2.0})
    )
    RADIANT_INTENSITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -3.0, _B.SOLID_ANGLE: -1.0})
    )
    REACTIVITY = ScalarVariableDescriptor(Dimensions({}))
    REFLECTIVITY = ScalarVariableDescriptor(Dimensions({}))
    REFRACTIVE_INDEX = ScalarVariableDescriptor(Dimensions({}))
    RELATIVE_ATOMIC_MASS = ScalarVariableDescriptor(Dimensions({}))
    RESILIENCE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    SCATTERING = ScalarVariableDescriptor(Dimensions({}))
    SEEBECK_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions(
            {
                _B.MASS: 1.0,
                _B.LENGTH: 2.0,
                _B.TIME: -3.0,
                _B.CURRENT: -1.0,
                _B.TEMPERATURE: -1.0,
            }
        )
    )
    SHEAR_MODULUS = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    SHEAR_STRENGTH = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    SKIN_FRICTION_COEFFICIENT = ScalarVariableDescriptor(Dimensions({}))
    SLIP = ScalarVariableDescriptor(Dimensions({}))
    SNAP = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0, _B.TIME: -4.0}))
    SOLID_ANGLE = ScalarVariableDescriptor(Dimensions({_B.SOLID_ANGLE: 1.0}))
    SOUND_REFLECTION = ScalarVariableDescriptor(Dimensions({}))
    SOUND_TRANSFER = ScalarVariableDescriptor(Dimensions({}))
    SPECIFIC_ACTIVITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: -1.0, _B.TIME: -1.0})
    )
    SPECIFIC_ANGULAR_MOMENTUM = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -1.0})
    )
    SPECIFIC_DISSIPATION_RATE = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    SPECIFIC_ENERGY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    SPECIFIC_ENTHALPY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    SPECIFIC_ENTROPY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0, _B.TEMPERATURE: -1.0})
    )
    SPECIFIC_HEAT_CAPACITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0, _B.TEMPERATURE: -1.0})
    )
    SPECIFIC_INTERNAL_ENERGY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    SPECIFIC_INTERNAL_SURFACE_AREA = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: -1.0})
    )
    SPECIFIC_MODULUS = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    SPECIFIC_STRENGTH = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    SPECIFIC_TOTAL_ENERGY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    SPECIFIC_TOTAL_ENTHALPY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    SPECIFIC_VOLUME = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 3.0, _B.MASS: -1.0})
    )
    SPECIFIC_WEIGHT = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -2.0, _B.TIME: -2.0})
    )
    SPECTRAL_INTENSITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.MASS: 1.0, _B.TIME: -3.0, _B.SOLID_ANGLE: -1.0})
    )
    SPECTRAL_IRRADIANCE = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: -1.0, _B.MASS: 1.0, _B.TIME: -3.0})
    )
    SPECTRAL_POWER = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.MASS: 1.0, _B.TIME: -3.0})
    )
    SPEED_OF_SOUND = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    SPIN_HALL_EFFECT = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 2.0, _B.TIME: -3.0, _B.CURRENT: -2.0})
    )
    STANDARD_ATOMIC_WEIGHT = ScalarVariableDescriptor(Dimensions({}))
    STATIC_PRESSURE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    STRAIN_RATE = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    STRESS = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    SURFACE_ENERGY = ScalarVariableDescriptor(Dimensions({_B.MASS: 1.0, _B.TIME: -2.0}))
    SURFACE_HEAT_FLUX = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -3.0})
    )
    SURFACE_HEAT_TRANSFER_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -3.0, _B.TEMPERATURE: -1.0})
    )
    SURFACE_NUSSELT_NUMBER = ScalarVariableDescriptor(Dimensions({}))
    SURFACE_ROUGHNESS = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0}))
    SURFACE_STANTON_NUMBER = ScalarVariableDescriptor(Dimensions({}))
    SURFACE_TENSION = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -2.0})
    )
    TANGENTIAL_VELOCITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    TEMPERATURE = ScalarVariableDescriptor(Dimensions({_B.TEMPERATURE: 1.0}))
    TEMPERATURE_DIFFERENCE = ScalarVariableDescriptor(
        Dimensions({_B.TEMPERATURE_DIFFERENCE: 1.0})
    )
    TEMPERATURE_GRADIENT = ScalarVariableDescriptor(
        Dimensions({_B.TEMPERATURE: 1.0, _B.LENGTH: -1.0})
    )
    TENSILE_STRENGTH = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    THERMAL_CONDUCTIVITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.MASS: 1.0, _B.TIME: -3.0, _B.TEMPERATURE: -1.0})
    )
    THERMAL_DIFFUSIVITY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -1.0})
    )
    THERMAL_EXPANSION_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions({_B.TEMPERATURE: -1.0})
    )
    THERMAL_RESISTANCE = ScalarVariableDescriptor(
        Dimensions({_B.TEMPERATURE: 1.0, _B.LENGTH: -2.0, _B.MASS: -1.0, _B.TIME: 3.0})
    )
    THIRD_ORDER_ELASTICITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    TIME = ScalarVariableDescriptor(Dimensions({_B.TIME: 1.0}))
    TORQUE = VectorVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    TORQUE_MAGNITUDE = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    TORQUE_X = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    TORQUE_Y = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    TORQUE_Z = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    TOTAL_ENERGY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    TOTAL_PRESSURE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    TOTAL_TEMPERATURE = ScalarVariableDescriptor(Dimensions({_B.TEMPERATURE: 1.0}))
    TOUGHNESS = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: -1.0, _B.MASS: 1.0, _B.TIME: -2.0})
    )
    TRANSMITTANCE = ScalarVariableDescriptor(Dimensions({}))
    TURBULENT_DISSIPATION_RATE = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: 3.0})
    )
    TURBULENT_INTENSITY = ScalarVariableDescriptor(Dimensions({}))
    TURBULENT_KINETIC_ENERGY = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 2.0, _B.TIME: -2.0})
    )
    TURBULENT_REYNOLDS_NUMBER = ScalarVariableDescriptor(Dimensions({}))
    TURBULENT_VISCOSITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -1.0})
    )
    TURBULENT_VISCOSITY_RATIO = ScalarVariableDescriptor(Dimensions({}))
    VAPOR_PRESSURE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    VELOCITY = VectorVariableDescriptor(Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0}))
    VELOCITY_MAGNITUDE = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0})
    )
    VELOCITY_X = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0}))
    VELOCITY_Y = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0}))
    VELOCITY_Z = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 1.0, _B.TIME: -1.0}))
    VISCOSITY = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -1.0})
    )
    VOLUME = ScalarVariableDescriptor(Dimensions({_B.LENGTH: 3.0}))
    VOLUMETRIC_FLOW = ScalarVariableDescriptor(
        Dimensions({_B.LENGTH: 3.0, _B.TIME: -1.0})
    )
    VOLUME_FRACTION = ScalarVariableDescriptor(Dimensions({}))
    VORTICITY = VectorVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    VORTICITY_MAGNITUDE = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    VORTICITY_X = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    VORTICITY_Y = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    VORTICITY_Z = ScalarVariableDescriptor(Dimensions({_B.TIME: -1.0}))
    WALL_ADJACENT_HEAT_TRANSFER_COEFFICIENT = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.TIME: -3.0, _B.TEMPERATURE: -1.0})
    )
    WALL_ADJACENT_TEMPERATURE = ScalarVariableDescriptor(
        Dimensions({_B.TEMPERATURE: 1.0})
    )
    WALL_SHEAR_STRESS = VectorVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    WALL_SHEAR_STRESS_MAGNITUDE = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    WALL_SHEAR_STRESS_X = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    WALL_SHEAR_STRESS_Y = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    WALL_SHEAR_STRESS_Z = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    WALL_TEMPERATURE = ScalarVariableDescriptor(Dimensions({_B.TEMPERATURE: 1.0}))
    WALL_TEMPERATURE_THIN = ScalarVariableDescriptor(Dimensions({_B.TEMPERATURE: 1.0}))
    WALL_Y_PLUS = ScalarVariableDescriptor(Dimensions({}))
    WALL_Y_STAR = ScalarVariableDescriptor(Dimensions({}))
    WAVENUMBER = ScalarVariableDescriptor(Dimensions({_B.LENGTH: -1.0}))
    YANK = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: 1.0, _B.TIME: -3.0})
    )
    YIELD_STRENGTH = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )
    YOUNG_MODULUS = ScalarVariableDescriptor(
        Dimensions({_B.MASS: 1.0, _B.LENGTH: -1.0, _B.TIME: -2.0})
    )

    @final
    class fluent:
        """Dictionary of variable descriptors for fluent-related quantities."""

        HELICITY = ScalarVariableDescriptor(_D.ACCELERATION)
        LAMBDA_2_CRITERION = ScalarVariableDescriptor(_D.TIME**-2)
        DENSITY_ALL = ScalarVariableDescriptor(_D.DENSITY)
        Y_PLUS_BASED_HEAT_TRANSFER_COEFFICIENT = ScalarVariableDescriptor(
            _D.POWER * _D.LENGTH**-2 * _D.TEMPERATURE**-1
        )
        TOTAL_ENTHALPY_DEVIATION = ScalarVariableDescriptor(_D.SPECIFIC_ENTHALPY)
        MASS_IMBALANCE = ScalarVariableDescriptor(_D.MASS * _D.TIME**-1)
        PRESSURE_HESSIAN_INDICATOR = ScalarVariableDescriptor(Dimensions())
        VELOCITY_ANGLE = ScalarVariableDescriptor(_D.ANGLE)
        DVELOCITY_DX = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DX_MAGNITUDE = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DX_X = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DX_Y = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DX_Z = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DY = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DY_MAGNITUDE = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DY_X = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DY_Y = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DY_Z = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DZ = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DZ_MAGNITUDE = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DZ_X = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DZ_Y = ScalarVariableDescriptor(_D.TIME**-1)
        DVELOCITY_DZ_Z = ScalarVariableDescriptor(_D.TIME**-1)
        VOLUME_FRACTION_PRIMARY_PHASE = ScalarVariableDescriptor(_D.VOLUME_FRACTION)
        VOLUME_FRACTION_SECONDARY_PHASE = ScalarVariableDescriptor(_D.VOLUME_FRACTION)

    @final
    class mesh:
        """Dictionary of variable descriptors for mesh-related quantities."""

        ANISOTROPIC_ADAPTION_CELLS = ScalarVariableDescriptor(Dimensions())
        BOUNDARY_CELL_DISTANCE = ScalarVariableDescriptor(Dimensions())
        BOUNDARY_LAYER_CELLS = ScalarVariableDescriptor(Dimensions())
        BOUNDARY_NORMAL_DISTANCE = ScalarVariableDescriptor(Dimensions())
        BOUNDARY_VOLUME_DISTANCE = ScalarVariableDescriptor(Dimensions())
        CELL_EQUIANGLE_SKEW = ScalarVariableDescriptor(Dimensions())
        CELL_EQUIVOLUME_SKEW = ScalarVariableDescriptor(Dimensions())
        CELL_PARENT_INDEX = ScalarVariableDescriptor(Dimensions())
        CELL_REFINE_LEVEL = ScalarVariableDescriptor(Dimensions())
        CELL_VOLUME = ScalarVariableDescriptor(_D.VOLUME)
        CELL_VOLUME_CHANGE = ScalarVariableDescriptor(Dimensions())
        ELEMENT_ASPECT_RATIO = ScalarVariableDescriptor(Dimensions())
        ELEMENT_WALL_DISTANCE = ScalarVariableDescriptor(_D.LENGTH)
        FACE_AREA_MAGNITUDE = ScalarVariableDescriptor(_D.AREA)
        FACE_HANDEDNESS = ScalarVariableDescriptor(Dimensions())
        INTERFACE_OVERLAP_FRACTION = ScalarVariableDescriptor(Dimensions())
        MARK_POOR_ELEMENTS = ScalarVariableDescriptor(Dimensions())
        SMOOTHED_CELL_REFINE_LEVEL = ScalarVariableDescriptor(Dimensions())
        X_FACE_AREA = ScalarVariableDescriptor(_D.AREA)
        Y_FACE_AREA = ScalarVariableDescriptor(_D.AREA)
        Z_FACE_AREA = ScalarVariableDescriptor(_D.AREA)
        ACTIVE_CELL_PARTITION = ScalarVariableDescriptor(Dimensions())
        CELL_ELEMENT_TYPE = ScalarVariableDescriptor(Dimensions())
        CELL_ID = ScalarVariableDescriptor(Dimensions())
        CELL_WEIGHT = ScalarVariableDescriptor(Dimensions())
        CELL_ZONE_INDEX = ScalarVariableDescriptor(Dimensions())
        CELL_ZONE_TYPE = ScalarVariableDescriptor(Dimensions())
        PARTITION_NEIGHBOURS = ScalarVariableDescriptor(Dimensions())
        STORED_CELL_PARTITION = ScalarVariableDescriptor(Dimensions())
        STORED_CELL_PARTITIION = ScalarVariableDescriptor(Dimensions())



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Defines the ConversionStrategy base class for translating `VariableDescriptor` objects.

Each concrete strategy maps `VariableDescriptor` instances to the string representations
required by a specific system.
"""

from abc import ABC, abstractmethod
from collections.abc import Mapping
import types

from typing_extensions import override

from ansys.units.variable_descriptor.variable_descriptor import VariableDescriptor


class ConversionStrategy(ABC):
    """
    Abstract base class for `VariableDescriptor` conversion strategies.

    This class defines the interface for all conversion strategies. Derived classes
    must implement the methods defined here to handle the conversion of `VariableDescriptor`
    objects to and from their string representations, as well as to check if a
    `VariableDescriptor` is supported.
    """

    @abstractmethod
    def to_string(self, variable: VariableDescriptor | str | None) -> str | None:
        """
        Convert a `VariableDescriptor` to its string representation.

        Parameters
        ----------
        variable : VariableDescriptor | str
            The `VariableDescriptor` to convert, or a string representation.

        Returns
        -------
        str
            The string representation of the `VariableDescriptor`.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in a derived class.
        """
        pass

    @abstractmethod
    def to_variable_descriptor(
        self, variable: VariableDescriptor | str
    ) -> VariableDescriptor | None:
        """
        Convert a string to its corresponding `VariableDescriptor`.

        Parameters
        ----------
        variable : VariableDescriptor | str
            The string representation to convert, or a `VariableDescriptor`.

        Returns
        -------
        VariableDescriptor
            The corresponding `VariableDescriptor` instance.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in a derived class.
        """
        pass

    @abstractmethod
    def supports(self, variable: VariableDescriptor) -> bool:
        """
        Check if the given `VariableDescriptor` is supported by the strategy.

        Parameters
        ----------
        variable : VariableDescriptor
            The `VariableDescriptor` to check.

        Returns
        -------
        bool
            `True` if the `VariableDescriptor` is supported, `False` otherwise.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in a derived class.
        """
        pass


class MappingConversionStrategy(ConversionStrategy, ABC):
    """
    Intermediate base class for implementing VariableDescriptor conversion strategies.

    This class simplifies the creation of concrete strategy classes by providing
    default implementations for common methods. Classes inheriting from this base
    class only need to define a `_mapping` dictionary that maps `VariableDescriptor`
    instances to their corresponding string representations.

    Attributes
    ----------
    _reverse_mapping : dict
        A lazily initialized reverse mapping of `_mapping`, used for converting
        strings back to `VariableDescriptor` instances.

    Methods
    -------
    to_string(variable: VariableDescriptor | str) -> str
        Converts a `VariableDescriptor` to its string representation. Raises a
        `ValueError` if the variable is not supported.
    to_variable_descriptor(variable: VariableDescriptor | str) -> VariableDescriptor
        Converts a string to its corresponding `VariableDescriptor`.
    supports(variable: VariableDescriptor) -> bool
        Checks if the given `VariableDescriptor` is supported by the strategy.

    Raises
    ------
    ValueError
        If a `VariableDescriptor` is not supported during conversion to a string.
    """

    _mapping: Mapping[VariableDescriptor, str]

    def __init__(self):
        """
        Initialize the MappingConversionStrategy.

        This constructor initializes the reverse mapping attribute to `None`.
        The reverse mapping is lazily initialized when accessed via the `_reverse_mapping` property.
        """
        self.__reverse_mapping: Mapping[str, VariableDescriptor] | None = None

    @property
    def _reverse_mapping(self):
        """
        Get the reverse mapping of `_mapping`.

        The reverse mapping is a dictionary that maps string representations
        back to their corresponding `VariableDescriptor` instances. It is lazily
        initialized on first access.

        Returns
        -------
        dict
            A dictionary mapping string representations to `VariableDescriptor` instances.
        """
        if self.__reverse_mapping is None:
            self.__reverse_mapping = {x: y for y, x in self._mapping.items()}
        return self.__reverse_mapping

    @override
    def to_string(self, variable: VariableDescriptor | str | None) -> str | None:
        """
        Convert a `VariableDescriptor` to its string representation.

        If the input is already a string, it is returned as-is. If the input
        is a `VariableDescriptor` and is supported by the strategy, its string
        representation is returned. Otherwise, a `ValueError` is raised.

        Parameters
        ----------
        variable : VariableDescriptor | str
            The `VariableDescriptor` to convert, or a string representation.

        Returns
        -------
        str
            The string representation of the `VariableDescriptor`.

        Raises
        ------
        ValueError
            If the `VariableDescriptor` is not supported by the strategy.
        """
        if isinstance(variable, (str, types.NoneType)):
            return variable
        if not self.supports(variable):
            raise ValueError(f"{variable.name} not supported.")
        return self._mapping[variable]

    @override
    def to_variable_descriptor(
        self, variable: VariableDescriptor | str
    ) -> VariableDescriptor | None:
        """
        Convert a string to its corresponding `VariableDescriptor`.

        If the input is already a `VariableDescriptor`, it is returned as-is.
        Otherwise, the string is converted to a `VariableDescriptor` using the
        reverse mapping.

        Parameters
        ----------
        variable : VariableDescriptor | str
            The string representation to convert, or a `VariableDescriptor`.

        Returns
        -------
        VariableDescriptor
            The corresponding `VariableDescriptor` instance, or `None` if the
            string is not found in the reverse mapping.
        """
        if isinstance(variable, VariableDescriptor):
            return variable
        return self._reverse_mapping.get(variable)

    @override
    def supports(self, variable: VariableDescriptor) -> bool:
        """
        Check if the given `VariableDescriptor` is supported by the strategy.

        Parameters
        ----------
        variable : VariableDescriptor
            The `VariableDescriptor` to check.

        Returns
        -------
        bool
            `True` if the `VariableDescriptor` is supported, `False` otherwise.
        """
        return variable in self._mapping



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Pandas extension for ansys-units.

This module provides pandas ExtensionDtype and ExtensionArray implementations for unit-
aware DataFrames and Series.
"""

# flake8: noqa: E501

import numbers
import operator
import re
from typing import Any, Callable

import numpy as np
import pandas as pd
from pandas.api.extensions import (
    ExtensionArray,
    ExtensionDtype,
    register_dataframe_accessor,
    register_extension_dtype,
    register_series_accessor,
    take,
)
from pandas.api.indexers import (
    check_array_indexer,
)
from pandas.api.types import (
    is_integer,
    is_object_dtype,
    is_string_dtype,
)
from typing_extensions import override

from ansys.units.quantity import Quantity, Vector
from ansys.units.unit import Unit

# Default subdtype for numeric data
DEFAULT_SUBDTYPE = "Float64"


@register_extension_dtype
class QuantityDtype(ExtensionDtype):
    """
    An ExtensionDtype for holding unit-aware Quantity data.

    Parameters
    ----------
    units : str or Unit
        The units for the quantities in this dtype.
    subdtype : str or dtype, optional
        The underlying numpy dtype for numeric values.

    Examples
    --------
    >>> import pandas as pd
    >>> from ansys.units.extensions.pandas_extension import QuantityDtype
    >>> dtype = QuantityDtype("m")
    >>> series = pd.Series([1.0, 2.0, 3.0], dtype=dtype)
    """

    type: "type[Quantity]" = Quantity
    units: Unit | None = None
    subdtype: Any | None = None
    _metadata: tuple[str, ...] = ("units", "subdtype")
    _match: re.Pattern[str] = re.compile(
        r"quantity\[(?P<units>.+)\](?:\[(?P<subdtype>.+)\])?"
    )
    _cache: dict[tuple[str, Any], "QuantityDtype"] = {}

    def __new__(
        cls, units: str | Unit | None = None, subdtype: Any | None = None
    ) -> "QuantityDtype":
        """
        Create a new QuantityDtype.

        Parameters
        ----------
        units : str or Unit, optional
            Units for this dtype.
        subdtype : str or dtype, optional
            Underlying dtype for numeric values.
        """
        if isinstance(units, QuantityDtype):
            return units

        if units is None:
            # Empty constructor for pickle compatibility
            return object.__new__(cls)

        # Parse units if string
        if isinstance(units, str):
            # Check if it's a full dtype string like "quantity[m][float64]"
            if units.startswith("quantity["):
                m = cls._match.search(units)
                if m:
                    units = m.group("units")
                    subdtype = subdtype or m.group("subdtype") or DEFAULT_SUBDTYPE

        # Convert string to Unit
        if not isinstance(units, Unit):
            units = Unit(units)

        if subdtype is None:
            subdtype = DEFAULT_SUBDTYPE

        # Use cached dtype if available
        try:
            return cls._cache[(str(units), subdtype)]
        except KeyError:
            u = object.__new__(cls)
            u.units = units
            u.subdtype = subdtype
            cls._cache[(str(units), subdtype)] = u
            return u

    @property
    def _is_numeric(self) -> bool:
        """Whether this dtype is numeric."""
        return True

    @override
    @classmethod
    def construct_from_string(cls, string: str) -> "QuantityDtype":
        """
        Construct dtype from string.

        Parameters
        ----------
        string : str
            String representation of the dtype.

        Returns
        -------
        QuantityDtype
        """
        if string.startswith("quantity["):
            return cls(string)

        # Try to interpret as unit string
        try:
            return cls(units=string)
        except Exception:
            raise TypeError(f"Cannot construct a 'QuantityDtype' from '{string}'")

    @property
    @override
    def name(self) -> str:
        """String representation of the dtype."""
        if self.subdtype and self.subdtype != DEFAULT_SUBDTYPE:
            return f"quantity[{self.units}][{self.subdtype}]"
        return f"quantity[{self.units}]"

    @property
    @override
    def na_value(self) -> Quantity:
        """The NA value for this dtype."""
        return Quantity(np.nan, str(self.units))

    @override
    def __hash__(self) -> int:
        """Hash the dtype."""
        return hash((str(self.units), str(self.subdtype)))

    @override
    def __eq__(self, other: Any) -> bool:
        """Check equality with another dtype."""
        if isinstance(other, str):
            try:
                other = QuantityDtype(other)
            except (ValueError, TypeError):
                return False

        if not isinstance(other, QuantityDtype):
            return False

        return str(self.units) == str(other.units) and self.subdtype == other.subdtype

    @override
    def __repr__(self) -> str:
        """String representation."""
        return self.name

    @override
    @classmethod
    def construct_array_type(cls) -> Any:
        """Return the array type associated with this dtype."""
        return QuantityArray


class QuantityArray(ExtensionArray):
    """
    An ExtensionArray for holding Quantity data.

    Parameters
    ----------
    values : array-like
        Numeric values for the quantities.
    dtype : QuantityDtype or str
        The dtype containing unit information.
    copy : bool, default False
        Whether to copy the input values.

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from ansys.units.extensions.pandas_extension import QuantityArray
    >>> arr = QuantityArray([1.0, 2.0, 3.0], dtype="quantity[m]")
    >>> series = pd.Series(arr)
    """

    _data: Any
    _dtype: QuantityDtype
    _can_hold_na: bool = True

    def __init__(
        self, values: Any, dtype: QuantityDtype | str | None = None, copy: bool = False
    ) -> None:
        """Initialize the QuantityArray."""
        # Parse dtype
        if dtype is None:
            if isinstance(values, QuantityArray):
                dtype = values._dtype
            elif isinstance(values, Quantity):
                dtype = QuantityDtype(units=values.units)
            else:
                raise ValueError("Must specify dtype when values are not Quantity")

        if not isinstance(dtype, QuantityDtype):
            dtype = QuantityDtype(dtype)

        self._dtype = dtype

        # Extract values from Quantity if needed
        if isinstance(values, Quantity):
            # Convert to target units
            values = values.to(str(dtype.units)).value
        elif isinstance(values, QuantityArray):
            values = values._data

        # Convert to pandas array with subdtype
        if not isinstance(values, ExtensionArray):
            values = pd.array(values, copy=copy, dtype=dtype.subdtype)

        self._data = values

    @property
    def dtype(self) -> QuantityDtype:
        """The dtype for this array."""
        return self._dtype

    def __len__(self) -> int:
        """Length of the array."""
        return len(self._data)

    def __getitem__(self, item: Any) -> Any:
        """
        Select subset of self.

        Parameters
        ----------
        item : int, slice, or ndarray
            Indexer for the array.

        Returns
        -------
        Quantity or QuantityArray
        """
        if is_integer(item):
            value = self._data[item]
            # Check if value is NaN/NA and return it directly
            if pd.isna(value):
                return value
            return Quantity(value, str(self._dtype.units))

        item: Any = check_array_indexer(self, item)
        return type(self)(self._data[item], self._dtype)

    def __setitem__(self, key: Any, value: Any) -> None:
        """
        Set values in the array.

        Parameters
        ----------
        key : int, slice, or ndarray
            Indexer for the array.
        value : scalar or array-like
            Value(s) to set.
        """
        # Convert Quantity to magnitude in our units
        if isinstance(value, Quantity):
            value = value.to(str(self._dtype.units)).value

        self._data[key] = value

    def isna(self) -> np.ndarray[Any, Any]:
        """
        Boolean indicator of missing values.

        Returns
        -------
        np.ndarray
        """
        return np.asarray(self._data.isna())

    def take(
        self, indices: Any, allow_fill: bool = False, fill_value: Any = None
    ) -> "QuantityArray":
        """
        Take elements from array.

        Parameters
        ----------
        indices : sequence of int
            Indices to take.
        allow_fill : bool, default False
            Whether to allow filling with fill_value.
        fill_value : any, optional
            Value to use for missing indices.

        Returns
        -------
        QuantityArray
        """
        if isinstance(fill_value, Quantity):
            fill_value = fill_value.to(str(self._dtype.units)).value

        result: np.ndarray[Any, Any] | ExtensionArray = take(
            self._data, indices, fill_value=fill_value, allow_fill=allow_fill
        )
        return QuantityArray(result, dtype=self.dtype)

    def copy(self, deep: bool = False) -> "QuantityArray":
        """
        Copy the array.

        Parameters
        ----------
        deep : bool, default False
            Whether to deep copy the underlying data.

        Returns
        -------
        QuantityArray
        """
        data = self._data
        if deep:
            data = data.copy()

        return type(self)(data, dtype=self.dtype)

    @classmethod
    def _concat_same_type(cls, to_concat: list["QuantityArray"]) -> "QuantityArray":
        """
        Concatenate multiple arrays of this type.

        Parameters
        ----------
        to_concat : list of QuantityArray

        Returns
        -------
        QuantityArray
        """
        output_units: Unit | None = to_concat[0]._dtype.units

        data: list[Any] = []
        for arr in to_concat:
            # Convert all arrays to output units
            converted: QuantityArray = arr.to(str(output_units))
            data.append(np.atleast_1d(converted._data))

        return cls(np.concatenate(data), to_concat[0].dtype)

    @classmethod
    def _from_sequence(
        cls, scalars: Any, dtype: QuantityDtype | None = None, copy: bool = False
    ) -> "QuantityArray":
        """
        Construct array from sequence of scalars.

        Parameters
        ----------
        scalars : sequence
            Sequence of values.
        dtype : QuantityDtype, optional
            Dtype for the array.
        copy : bool, default False
            Whether to copy the data.

        Returns
        -------
        QuantityArray
        """
        # Try to find a Quantity in scalars to infer units
        master_quantity: Quantity | None = None
        for item in scalars:
            if isinstance(item, Quantity):
                master_quantity = item
                break

        if dtype is None:
            if master_quantity is None:
                raise ValueError(
                    "Cannot infer dtype. No dtype specified and no Quantity found"
                )
            dtype = QuantityDtype(units=master_quantity.units)

        # Convert all Quantities to magnitudes in target units
        values: list[Any] = []
        for item in scalars:
            if isinstance(item, Quantity):
                values.append(item.to(str(dtype.units)).value)
            else:
                values.append(item)

        return cls(values, dtype=dtype, copy=copy)

    @classmethod
    def _from_factorized(
        cls, values: Any, original: "QuantityArray"
    ) -> "QuantityArray":
        """Reconstruct from factorized values."""
        return cls(values, dtype=original.dtype)

    def _values_for_factorize(self) -> tuple[np.ndarray[Any, Any], Quantity]:
        """Return values for factorization."""
        return np.asarray(self._data), self.dtype.na_value

    @property
    def quantity(self) -> Quantity:
        """Convert to single Quantity with array value."""
        return Quantity(np.asarray(self._data), str(self._dtype.units))

    def to(self, units: str | Unit) -> "QuantityArray":
        """
        Convert to different units.

        Parameters
        ----------
        units : str or Unit
            Target units.

        Returns
        -------
        QuantityArray
        """
        if not isinstance(units, Unit):
            units = Unit(units)

        # Convert the quantity
        converted: Quantity[float | Vector] = self.quantity.to(str(units))
        return QuantityArray(converted.value, dtype=QuantityDtype(units))

    def astype(self, dtype: Any, copy: bool = True) -> Any:
        """
        Cast to different dtype.

        Parameters
        ----------
        dtype : str or dtype
            Target dtype.
        copy : bool, default True
            Whether to copy the data.

        Returns
        -------
        array
        """
        if isinstance(dtype, str) and dtype.startswith("quantity["):
            dtype = QuantityDtype(dtype)

        if isinstance(dtype, QuantityDtype):
            if dtype == self._dtype and not copy:
                return self
            return self.to(str(dtype.units))

        if is_object_dtype(dtype):
            return np.array([self[i] for i in range(len(self))], dtype=object)

        if is_string_dtype(dtype):
            return pd.Series([str(self[i]) for i in range(len(self))], dtype=dtype)

        return pd.array(self._data, dtype=dtype, copy=copy)

    def __array__(
        self, dtype: Any | None = None, copy: bool = False
    ) -> np.ndarray[Any, Any]:
        """Convert to numpy array."""
        if dtype is None or is_object_dtype(dtype):
            # Return array of Quantity objects
            return np.array([self[i] for i in range(len(self))], dtype=object)
        return np.array(self._data, dtype=dtype, copy=copy)

    @staticmethod
    def _create_arithmetic_method(
        op: Callable[..., Any], reflected: bool = False
    ) -> Callable[..., Any]:
        """
        Create arithmetic method.

        Parameters
        ----------
        op : Callable
            The binary operator to apply.
        reflected : bool, optional
            If ``True``, reverse operand order so that ``op(rhs, lhs)`` is
            computed instead of ``op(lhs, rhs)``.  Required for
            ``__rsub__`` and ``__rtruediv__`` where the array is the *right*
            operand of the original expression.
        """

        def method(self: "QuantityArray", other: Any) -> Any:
            if isinstance(other, (pd.Series, pd.DataFrame, pd.Index)):
                return NotImplemented

            # Convert to quantities
            lhs: Quantity[float | Vector] = self.quantity
            rhs: Any

            if isinstance(other, QuantityArray):
                rhs = other.quantity
            elif isinstance(other, Quantity):
                rhs = other
            elif isinstance(other, (numbers.Number, np.ndarray)):
                rhs = other
            else:
                return NotImplemented

            # Perform operation — reflected ops swap operand order so that
            # e.g. ``x - array`` computes ``op(x, array)`` not ``op(array, x)``.
            result = op(rhs, lhs) if reflected else op(lhs, rhs)

            # Convert back to array
            if isinstance(result, Quantity):
                return QuantityArray._from_sequence([result])
            return NotImplemented

        return method

    @staticmethod
    def _create_comparison_method(op: Callable[..., Any]) -> Callable[..., Any]:
        """Create comparison method."""

        def method(self: "QuantityArray", other: Any) -> Any:
            if isinstance(other, (pd.Series, pd.DataFrame, pd.Index)):
                return NotImplemented

            lhs: Quantity[float | Vector] = self.quantity

            if isinstance(other, QuantityArray):
                rhs: Quantity[float | Vector] = other.quantity
            elif isinstance(other, Quantity):
                rhs = other
            else:
                rhs = other

            result = op(lhs, rhs)
            return result

        return method

    def _reduce(self, name: str, *, skipna: bool = True, **kwargs: Any) -> "Quantity":
        functions: dict[str, Callable[..., Any]] = {
            "sum": np.nansum,
            "mean": np.nanmean,
            "min": np.nanmin,
            "max": np.nanmax,
            "std": np.nanstd,
            "var": np.nanvar,
        }

        if name not in functions:
            raise TypeError(f"Cannot perform {name} with dtype {self.dtype}")

        data = np.asarray(self._data)

        # Remove pandas-specific kwargs that NumPy does not support
        np_kwargs = {k: v for k, v in kwargs.items() if k not in {"min_count"}}

        if skipna:
            result = functions[name](data, **np_kwargs)
        else:
            result = getattr(np, name)(data, **np_kwargs)

        # Adjust units for variance
        if name == "var":
            return Quantity(result, f"({self._dtype.units})**2")
        return Quantity(result, str(self._dtype.units))

    # Ensure pandas / numpy defer to us for ops
    __array_priority__: int = 1000

    # Arithmetic operators
    __add__: Callable[[Any, Any], Any] = _create_arithmetic_method(operator.add)
    __sub__: Callable[[Any, Any], Any] = _create_arithmetic_method(operator.sub)
    __mul__: Callable[[Any, Any], Any] = _create_arithmetic_method(operator.mul)
    __truediv__: Callable[[Any, Any], Any] = _create_arithmetic_method(operator.truediv)

    # Reverse ops
    __radd__: Callable[[Any, Any], Any] = _create_arithmetic_method(operator.add)
    __rsub__: Callable[[Any, Any], Any] = _create_arithmetic_method(
        operator.sub, reflected=True
    )
    __rmul__: Callable[[Any, Any], Any] = _create_arithmetic_method(operator.mul)
    __rtruediv__: Callable[[Any, Any], Any] = _create_arithmetic_method(
        operator.truediv, reflected=True
    )

    # Comparisons
    __eq__: Callable[[Any, Any], Any] = _create_comparison_method(operator.eq)
    __ne__: Callable[[Any, Any], Any] = _create_comparison_method(operator.ne)
    __lt__: Callable[[Any, Any], Any] = _create_comparison_method(operator.lt)
    __le__: Callable[[Any, Any], Any] = _create_comparison_method(operator.le)
    __gt__: Callable[[Any, Any], Any] = _create_comparison_method(operator.gt)
    __ge__: Callable[[Any, Any], Any] = _create_comparison_method(operator.ge)


@register_series_accessor("units")
class UnitsSeriesAccessor:
    """
    Accessor for unit-aware operations on Series.

    Examples
    --------
    >>> import pandas as pd
    >>> from ansys.units.extensions.pandas_extension import QuantityDtype
    >>> series = pd.Series([1.0, 2.0, 3.0], dtype=QuantityDtype("m"))
    >>> series.units.to("ft")
    """

    def __init__(self, pandas_obj: Any) -> None:
        """Initialize the accessor."""
        self._obj = pandas_obj
        if not isinstance(pandas_obj.dtype, QuantityDtype):
            raise AttributeError(
                "Series must have QuantityDtype to use .units accessor"
            )

    def to(self, units: str | Unit) -> Any:
        """
        Convert to different units.

        Parameters
        ----------
        units : str or Unit
            Target units.

        Returns
        -------
        pd.Series
        """
        if not isinstance(self._obj.array, QuantityArray):
            raise TypeError("Series must contain QuantityArray")

        converted: QuantityArray = self._obj.array.to(units)
        return pd.Series(converted, index=self._obj.index, name=self._obj.name)

    @property
    def quantity(self) -> Quantity:
        """Get the underlying Quantity."""
        if not isinstance(self._obj.array, QuantityArray):
            raise TypeError("Series must contain QuantityArray")
        return self._obj.array.quantity


@register_dataframe_accessor("units")
class UnitsDataFrameAccessor:
    """
    Accessor for unit-aware operations on DataFrames.

    Examples
    --------
    >>> import pandas as pd
    >>> from ansys.units.extensions.pandas_extension import QuantityDtype
    >>> df = pd.DataFrame({
    ...     "length": pd.Series([1.0, 2.0], dtype=QuantityDtype("m")),
    ...     "mass": pd.Series([10.0, 20.0], dtype=QuantityDtype("kg"))
    ... })
    >>> df.units.summary()
    """

    def __init__(self, pandas_obj: Any) -> None:
        """Initialize the accessor."""
        self._obj = pandas_obj

    def summary(self) -> dict[str, str]:
        """
        Get summary of units in the DataFrame.

        Returns
        -------
        dict
            Dictionary mapping column names to their units.
        """
        summary: dict[str, str] = {}
        for col in self._obj.columns:
            if isinstance(self._obj[col].dtype, QuantityDtype):
                summary[col] = str(self._obj[col].dtype.units)
        return summary

    def to(self, units_map: dict[str, str | Unit]) -> Any:
        """
        Convert columns to different units.

        Parameters
        ----------
        units_map : dict
            Dictionary mapping column names to target units.

        Returns
        -------
        pd.DataFrame
        """
        df = self._obj.copy()
        for col, units in units_map.items():
            if col in df.columns and isinstance(df[col].dtype, QuantityDtype):
                df[col] = df[col].units.to(units)
        return df



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Mapping, Sequence
import functools
import os
from typing import TYPE_CHECKING, Any, overload

from ansys.units._constants import (
    _base_units,
    _BaseUnitInfo,
    _derived_units,
    _DerivedUnitInfo,
    _multipliers,
    _quantity_units_table,
)
from ansys.units.base_dimensions import BaseDimensions
from ansys.units.dimensions import Dimensions
from ansys.units.quantity_tables.keys import QuantityKey
from ansys.units.systems import UnitSystem

if TYPE_CHECKING:
    import numpy as np
    import numpy.typing as npt

    from ansys.units.quantity import Quantity, ValT, Vector


@functools.cache
def _get_quantity_and_array():
    from ansys.units.quantity import Quantity, _array

    return Quantity, _array


class Unit:
    """
    A class containing the string name and dimensions of a unit.

    Parameters
    ----------
    units: str, optional
        Name of the unit or string chain of combined units
    config: dict, optional
        dictionary of unit properties
    dimensions: Dimensions, optional
        An instance of the Dimensions class.
    system: str, optional
        Define the unit system for base units of dimension,
        default is SI.
    table: dict, optional
        A dictionary of api table keys from the cfg.yaml and exponent values.
    copy_from: Unit, optional
        A previous instance of Unit.

    Attributes
    ----------
    name
    si_units
    si_scaling_factor
    si_offset
    dimensions

    See Also
    --------
    :mod:`ansys.units.common`
        A collection of predefined units for use.

    Examples
    --------
    >>> from ansys.units import Unit, Quantity
    >>> fps = Unit("ft s^-1")
    >>> fps.name
    'ft s^-1'
    >>> fps.dimensions
    {'LENGTH': 1.0, 'TIME': -1.0}
    >>> speed = Quantity(value=5, units=fps)
    >>> speed
    Quantity(5.0, "ft s^-1")
    """

    def __init__(
        self,
        units: str | None = None,
        config: Mapping[str, Any] | None = None,
        dimensions: Dimensions | None = None,
        system: UnitSystem | str = None,
        table: Mapping[QuantityKey, float] | None = None,
        copy_from: Unit | None = None,
    ):
        if copy_from:
            if (units) and units != copy_from.name:
                raise InconsistentDimensions()
            units = copy_from.name

        if table:
            units = _table_to_units(table=table)

        if units:
            self._name = _condense(units)
            _dimensions = _units_to_dim(units=units)
            self._dimensions = Dimensions(_dimensions)
            if dimensions and self._dimensions != dimensions:
                raise InconsistentDimensions()
            self._dimensions = self._remove_angle_as_dim(self._dimensions)

        elif dimensions:
            self._dimensions = dimensions
            self._name = _dim_to_units(dimensions=dimensions, system=system)
            self._dimensions = self._remove_angle_as_dim(self._dimensions)
        else:
            self._name = ""
            self._dimensions = Dimensions()

        if not config:
            config = _get_config(name=self._name)
        if config:
            for key in config:
                setattr(self, f"_{key}", config[key])

        self._si_units, self._si_scaling_factor, self._si_offset = _si_data(
            units=self.name
        )

    def _remove_angle_as_dim(self, dimensions):
        if not os.getenv("PYANSYS_UNITS_ANGLE_AS_DIMENSION", None):
            if BaseDimensions.ANGLE in dimensions._dimensions:
                del dimensions._dimensions[BaseDimensions.ANGLE]
            if BaseDimensions.SOLID_ANGLE in dimensions._dimensions:
                del dimensions._dimensions[BaseDimensions.SOLID_ANGLE]
        return dimensions

    def _to_string(self):
        """
        Creates a string representation of the unit.

        Returns
        -------
        str
            A string version of the unit.
        """
        returned_string = ""
        attrs = self.__dict__
        for key in attrs:
            returned_string += f"{key}: {attrs[key]}\n"
        return returned_string

    def _new_units(self, value, op):
        """
        Generate a new units instance depending on mathematical operation.

        Parameters
        ----------
        value : int | float | Unit
            The given operand.
        op : str
            '*', '/', '**'.

        Returns
        -------
        Unit
            New unit instance.
        """
        new_units = ""
        if op == "**":
            for term in self.name.split(" "):
                multiplier, base, exponent = _filter_unit_term(term)
                exponent *= value
                new_units += f"{multiplier}{base}^{exponent} "
        if op == "/":
            new_units = self.name
            for term in value.name.split(" "):
                multiplier, base, exponent = _filter_unit_term(term)
                new_units += f" {multiplier}{base}^{exponent * -1}"
        if op == "*":
            new_units = f"{self.name} {value.name}"
        return Unit(_condense(new_units))

    def compatible_units(self) -> set[str]:
        """
        Get all units with the same dimensions.

        Returns
        -------
        set
            A set of unit objects.
        """
        compatible_units = set()
        for unit_name in {**_base_units, **_derived_units}.keys():
            unit = Unit(units=unit_name)
            if self.dimensions == unit.dimensions:
                compatible_units.add(unit.name)
        for unit_name in _quantity_units_table:
            unit = Unit(table={unit_name: 1})
            if self.dimensions == unit.dimensions:
                compatible_units.add(unit.name)
        compatible_units.discard(self.name)
        return compatible_units

    def _temp_precheck(self, other_unit, op: str = "+") -> tuple["Unit", "Unit"] | None:
        """
        Validate units for temperature differences.

        Parameters
        ----------
        other_unit : Unit
            Unit for comparison against current unit.
        op : str, optional
            Operation conducted on the units.

        Returns
        -------
        Unit, Unit | None
            Two unit objects for a quantity of temperature difference or temperature.

        Raises
        ------
        ProhibitedTemperatureOperation
            Cannot add two absolute temperatures.
        IncorrectUnits
            Cannot add or subtract different units.
        """
        base = BaseDimensions
        dims = Dimensions
        temp = dims(dimensions={base.TEMPERATURE: 1})
        delta_temp = dims(dimensions={base.TEMPERATURE_DIFFERENCE: 1})
        if self.dimensions == other_unit.dimensions == temp and op == "+":
            raise ProhibitedTemperatureOperation()

        # Checks to make sure they are both temperatures.
        if (self.dimensions and other_unit.dimensions) in (temp, delta_temp):
            unit_name = self.name.removeprefix("delta_")
            relative = Unit(f"delta_{unit_name}")
            absolute = Unit(unit_name)

            if self.dimensions != other_unit.dimensions:
                # Removes the delta_ prefix if there is one.
                return absolute, relative

            if self.dimensions == other_unit.dimensions == temp and op == "-":
                return relative, absolute

        if self.dimensions != other_unit.dimensions:
            raise IncorrectUnits(self, other_unit)

    @property
    def name(self) -> str:
        """The unit string."""
        return self._name

    @property
    def si_units(self) -> str:
        """The unit string in SI units."""
        return self._si_units

    @property
    def si_scaling_factor(self) -> float:
        """The scaling factor used to convert to SI units."""
        return self._si_scaling_factor

    @property
    def si_offset(self) -> float:
        """The offset used to convert to SI units."""
        return self._si_offset

    @property
    def dimensions(self) -> Dimensions:
        """The units base dimensions."""
        return self._dimensions

    def convert(self, system: UnitSystem) -> Unit:
        """
        Convert a unit into the unit system.

        Parameters
        ----------
        system : UnitSystem
            Unit system to convert to.

        Returns
        -------
        Unit
            Unit object converted into the unit system.

        Examples
        --------
        >>> ur = UnitRegistry()
        >>> speed_si = Unit(units= ur.m / ur.s)
        >>> bt = UnitSystem(system="BT")
        >>> speed_bt = speed_si.convert(bt)
        """

        return Unit(dimensions=self.dimensions, system=system)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Unit({self.name!r})"

    def __add__(self, value) -> tuple["Unit", "Unit"]:
        return self._temp_precheck(value)

    def __mul__(self, value: "Unit") -> "Unit":
        if isinstance(value, Unit):
            new_units = f"{self.name} {value.name}"
            return Unit(_condense(new_units))

        else:
            return NotImplemented

    @overload
    def __rmul__(
        self, value: Sequence[float] | Vector
    ) -> "Quantity[npt.NDArray[np.floating]]": ...

    @overload
    def __rmul__(self, value: float) -> "Quantity[float]": ...

    def __rmul__(self, value: Sequence[float] | Vector | float) -> "Quantity[ValT]":
        Quantity, _array = _get_quantity_and_array()
        if isinstance(value, (float, int)) or (
            _array and isinstance(value, (Sequence, _array.ndarray))
        ):
            return Quantity(value=value, units=self)
        return NotImplemented  # other cases should have already handled this so nothing to do

    def __sub__(self, value: "Unit") -> tuple["Unit", "Unit"]:
        return self._temp_precheck(value, op="-")

    def __truediv__(self, value: Unit) -> Unit:
        if not isinstance(value, Unit):
            return NotImplemented
        new_units = ""
        new_units = self.name
        for term in value.name.split(" "):
            multiplier, base, exponent = _filter_unit_term(term)
            new_units += f" {multiplier}{base}^{exponent * -1}"
        return Unit(_condense(new_units))

    @overload
    def __rtruediv__(
        self, value: Sequence[float] | Vector
    ) -> "Quantity[npt.NDArray[np.floating]]": ...

    @overload
    def __rtruediv__(self, value: float) -> "Quantity[float]": ...

    def __rtruediv__(self, value: Sequence[float] | Vector | float) -> "Quantity[ValT]":
        Quantity, _array = _get_quantity_and_array()
        if isinstance(value, (float, int)) or (
            _array and isinstance(value, (Sequence, _array.ndarray))
        ):
            return Quantity(value=value, units=self**-1)
        return NotImplemented  # other cases should have already handled this so nothing to do

    def __pow__(self, value: "Unit" | float) -> "Unit":
        new_units = ""
        for term in self.name.split(" "):
            multiplier, base, exponent = _filter_unit_term(term)
            exponent *= value
            new_units += f"{multiplier}{base}^{exponent} "

        return Unit(_condense(new_units))

    def __eq__(self, other_unit: object) -> bool:
        if not isinstance(other_unit, Unit) and self.name:
            return False
        if isinstance(other_unit, Unit):
            checks = [
                self.dimensions == other_unit.dimensions,
                self.si_offset == other_unit.si_offset,
                self.si_scaling_factor == other_unit.si_scaling_factor,
            ]
            return all(checks)
        return True


def _get_config(name: str) -> _BaseUnitInfo | _DerivedUnitInfo | None:
    """
    Retrieve unit configuration from '_base_units' or '_derived_units'.

    Parameters
    ----------
    name : str
        Unit string.
    Returns
    -------
    dict
        Dictionary of extra unit information.
    """
    if name in _base_units:
        return _base_units[name]
    if name in _derived_units:
        return _derived_units[name]


def _dim_to_units(
    system: UnitSystem,
    dimensions: Dimensions,
) -> str:
    """
    Convert a dimensions list into a unit string.

    Parameters
    ----------
    dimensions : Dimensions object
        Instance of Dimension class.

        system : UnitSystem object, optional
            Unit system for dimensions list.
            Default is SI units.

        Returns
        -------
        str
            Unit string.
    """
    if not system:
        system = UnitSystem()

    units = ""

    for key, value in dimensions:
        if value == 1:
            units += f"{getattr(system, key.name)} "
        elif value != 0.0:
            value = int(value) if value % 1 == 0 else value
            units += f"{getattr(system, key.name)}^{value} "

    return units.strip()


def _units_to_dim(
    units: str,
    exponent: float = 1.0,
    dimensions: Mapping[
        BaseDimensions, float
    ] = {},  # pyright: ignore[reportCallInDefaultInitializer]
) -> dict[BaseDimensions, float]:
    """
    Convert a unit string into a Dimensions instance.

    Parameters
    ----------
    units : str
        Unit string.
    Returns
    -------
    dict
        Dimensions dictionary
    """
    dimensions = dict(dimensions)
    # Split unit string into terms and parse data associated with individual terms
    for term in units.split(" "):
        _, unit_term, unit_term_exponent = _filter_unit_term(term)
        unit_term_exponent *= exponent
        # retrieve data associated with base unit
        if unit_term in _base_units:
            idx = _base_units[unit_term]["type"]

            if BaseDimensions[idx] in dimensions:
                dimensions[BaseDimensions[idx]] += unit_term_exponent
            else:
                dimensions[BaseDimensions[idx]] = unit_term_exponent
        # Retrieve derived unit composition unit string and SI factor.
        elif unit_term in _derived_units:
            # Recursively parse composition unit string

            dimensions = _units_to_dim(
                units=_derived_units[unit_term]["composition"],
                exponent=unit_term_exponent,
                dimensions=dimensions,
            )
        elif _:
            raise UnconfiguredUnit(_)

    return dimensions


def _table_to_units(table: Mapping[QuantityKey, int]) -> str:
    """
    Convert a quantity table item into a unit string.

    Parameters
    ----------
    table : dict[str, int]
        Quantity table item to convert to a Unit.

    Returns
    -------
    Unit
        Unit object representation of a quantity table item.
    """
    for key in table.keys():
        if key not in _quantity_units_table:
            raise UnknownTableItem(key)

    base_unit = ""

    for key, value in table.items():
        terms = _quantity_units_table[key]
        for term in terms.split(" "):
            multiplier, base, exponent = _filter_unit_term(term)

            base_unit += f"{multiplier}{base}^{exponent * value} "

    return _condense(base_unit)


def _multiplier_check(unit_term: str) -> bool:
    """
    Check if a unit term contains a multiplier.

    Parameters
    ----------
    unit_term : str
        Unit term of the unit string.

    Returns
    -------
    bool
        ``True`` if the unit term contains a multiplier, ``False`` otherwise.
    """
    # Check if the unit term is not an existing base or derived unit.
    return unit_term and not (
        (unit_term in _base_units) or (unit_term in _derived_units)
    )


def _si_map(unit_term: str) -> str:
    """
    Convert a unit into its SI equivalent.

    Parameters
    ----------
    unit_term : str
        Unit term of the unit string.

    Returns
    -------
    str
        SI unit equivalent.
    """
    # Retrieve type associated with unit term
    unit_term_type = _base_units[unit_term]["type"]

    # Find SI unit with same type as unit term
    for term, term_info in _base_units.items():
        if (
            term_info["type"] == unit_term_type
            and term_info["si_scaling_factor"] == 1.0
        ):
            return term


def _condense(units: str) -> str:
    """
    Condense a unit string by collecting like terms.

    Parameters
    ----------
    units : str
        Unit string to simplify.

    Returns
    -------
    str
        Simplified unit string.
    """
    terms_and_exponents = {}
    units = units.strip()

    # Split unit string into terms and parse data associated with individual terms
    for term in units.split(" "):
        multiplier, unit_term, unit_term_exponent = _filter_unit_term(term)
        full_term = f"{multiplier}{unit_term}"
        if full_term in terms_and_exponents:
            terms_and_exponents[full_term] += unit_term_exponent
        else:
            terms_and_exponents[full_term] = unit_term_exponent
    units = ""
    # Concatenate unit string
    for term, exponent in terms_and_exponents.items():
        if not (exponent):
            continue
        if exponent == 1.0:
            units += f"{term} "
        else:
            exponent = int(exponent) if exponent % 1 == 0 else exponent
            units += f"{term}^{exponent} "

    return units.rstrip()


def _filter_unit_term(unit_term: str) -> tuple[str, str, float]:
    """
    Separate multiplier, base, and exponent from a unit term.

    Parameters
    ----------
    unit_term : str
        Unit term of the unit string.

    Returns
    -------
    tuple
        Tuple containing the multiplier, base, and exponent of the unit term.
    """
    multiplier = ""
    exponent = 1.0

    # strip exponent from unit term
    if "^" in unit_term:
        exponent = float(unit_term[unit_term.index("^") + 1 :])
        unit_term = unit_term[: unit_term.index("^")]

    base = unit_term

    # strip multiplier and base from unit term
    has_multiplier = _multiplier_check(unit_term)
    if has_multiplier:
        for mult in _multipliers:
            if unit_term.startswith(mult):
                if not _multiplier_check(unit_term[len(mult) :]):
                    multiplier = mult
                    base = unit_term[len(mult) :]
                    break

    # if we thought it had a multiplier, that's just because the string wasn't
    # a known unit on its own. So if we can't actually find its multiplier then
    # this string is an invalid unit string
    if has_multiplier and not multiplier:
        raise UnconfiguredUnit(unit_term)
    return multiplier, base, exponent


def _si_data(
    units: str,
    exponent: float = None,
    si_units: str = None,
    si_scaling_factor: float = None,
) -> tuple:
    """
    Compute the SI unit string, SI scaling factor, and SI offset.

    Parameters
    ----------
    units : str
        Unit string representation of the quantity.
    exponent : float, None
        exponent of the unit string.
    si_units : str, None
        SI unit string representation of the quantity.
    si_scaling_factor : float, None
        SI scaling factor of the unit string.

    Returns
    -------
    tuple
        Tuple containing the SI units, SI scaling factor, and SI offset.
    """
    # Initialize default values
    units = units or " "
    exponent = exponent or 1.0
    si_units = si_units or ""
    si_scaling_factor = si_scaling_factor or 1.0
    si_offset = _base_units[units]["si_offset"] if units in _base_units else 0.0

    # Split unit string into terms and parse data associated with individual terms
    for term in units.split(" "):
        unit_multiplier, unit_term, unit_term_exponent = _filter_unit_term(term)

        unit_term_exponent *= exponent

        si_scaling_factor *= (
            _multipliers[unit_multiplier] ** unit_term_exponent
            if unit_multiplier
            else 1.0
        )

        # Retrieve data associated with base unit
        if unit_term in _base_units:
            if unit_term_exponent == 1.0:
                si_units += f" {_si_map(unit_term)}"
            elif unit_term_exponent != 0.0:
                si_units += f" {_si_map(unit_term)}^{unit_term_exponent}"

            si_scaling_factor *= (
                _base_units[unit_term]["si_scaling_factor"] ** unit_term_exponent
            )

        # Retrieve derived unit composition unit string and SI scaling factor
        elif unit_term in _derived_units:
            si_scaling_factor *= (
                _derived_units[unit_term]["factor"] ** unit_term_exponent
            )

            # Recursively parse composition unit string
            si_units, si_scaling_factor, _ = _si_data(
                units=_derived_units[unit_term]["composition"],
                exponent=unit_term_exponent,
                si_units=si_units,
                si_scaling_factor=si_scaling_factor,
            )

    return _condense(si_units), si_scaling_factor, si_offset


class InconsistentDimensions(ValueError):
    """Raised when units have inconsistent base dimensions."""

    def __init__(self):
        super().__init__("Unit dimensions do not match given dimensions.")


class UnconfiguredUnit(ValueError):
    """Raised when the specified unit is unconfigured."""

    def __init__(self, unit):
        super().__init__(f"`{unit}` is an unconfigured unit.")


class IncorrectUnits(ValueError):
    """Raised when the specified units are incorrect."""

    def __init__(self, unit1, unit2):
        super().__init__(
            f"'{unit1.si_units}' and '{unit2.si_units}' must match for this operation."
        )


class ProhibitedTemperatureOperation(TypeError):
    """Raised when two absolute temperatures are added."""

    def __init__(self):
        super().__init__(
            "In order to add two temperature-based quantities, at least one must be a temperature difference."  # noqa: E501
        )


class UnknownTableItem(ValueError):
    """Raised when a quantity table item is undefined."""

    def __init__(self, item):
        super().__init__(f"`{item}` is not a valid quantity table item.")
